@@ -41,7 +41,9 @@ static char *filename;
 
 #ifdef LIBXML_CATALOG_ENABLED
 
+#ifndef XML_SGML_DEFAULT_CATALOG
 #define XML_SGML_DEFAULT_CATALOG "/etc/sgml/catalog"
+#endif
 
 /************************************************************************
  * 									*
@@ -72,17 +74,23 @@ xmlShellReadline(const char *prompt) {
     return (line_read);
 #else
     char line_read[501];
+    char *ret;
+    int len;
 
     if (prompt != NULL)
 	fprintf(stdout, "%s", prompt);
     if (!fgets(line_read, 500, stdin))
         return(NULL);
     line_read[500] = 0;
-    return(strdup(line_read));
+    len = strlen(line_read);
+    ret = (char *) malloc(len + 1);
+    if (ret != NULL) {
+	memcpy (ret, line_read, len + 1);
+    }
+    return(ret);
 #endif
 }
 
-
 static void usershell(void) {
     char *cmdline = NULL, *cur;
     int nbargs;
@@ -204,11 +212,17 @@ static void usershell(void) {
 	    }
 	} else if (!strcmp(command, "add")) {
 	    if (sgml) {
-		if (nbargs != 1) {
-		    printf("add requires 1 argument\n");
+		if ((nbargs != 3) && (nbargs != 2)) {
+		    printf("add requires 2 or 3 arguments\n");
 		} else {
-		    ret = xmlCatalogAdd(BAD_CAST "sgmlcatalog", NULL,
-			                BAD_CAST argv[0]);
+		    if (argv[2] == NULL)
+			ret = xmlCatalogAdd(BAD_CAST argv[0], NULL,
+					    BAD_CAST argv[1]);
+		    else
+			ret = xmlCatalogAdd(BAD_CAST argv[0], BAD_CAST argv[1],
+					    BAD_CAST argv[2]);
+		    if (ret != 0)
+			printf("add command failed\n");
 		}
 	    } else {
 		if ((nbargs != 3) && (nbargs != 2)) {
@@ -307,6 +321,7 @@ static void usage(const char *name) {
 int main(int argc, char **argv) {
     int i;
     int ret;
+    int exit_value = 0;
 
 
     if (argc <= 1) {
@@ -375,12 +390,10 @@ int main(int argc, char **argv) {
 	} else if (argv[i][0] == '-')
 	    continue;
 	filename = argv[i];
-	if (!sgml) {
 	    ret = xmlLoadCatalog(argv[i]);
 	    if ((ret < 0) && (create)) {
 		xmlCatalogAdd(BAD_CAST "catalog", BAD_CAST argv[i], NULL);
 	    }
-	}
 	break;
     }
 
@@ -400,7 +413,7 @@ int main(int argc, char **argv) {
 
 	    if (sgml) {
 		/*
-		 * Maintainance of SGML catalogs.
+		 * Maintenance of SGML catalogs.
 		 */
 		xmlCatalogPtr catal = NULL;
 		xmlCatalogPtr super = NULL;
@@ -424,9 +437,11 @@ int main(int argc, char **argv) {
 			ret = xmlACatalogRemove(catal, BAD_CAST argv[i + 2]);
 		    else
 			ret = -1;
-		    if (ret < 0)
-			fprintf(stderr, "Failed to removed entry from %s\n",
+		    if (ret < 0) {
+			fprintf(stderr, "Failed to remove entry from %s\n",
 				argv[i + 1]);
+			exit_value = 1;
+		    }
 		    if ((noout) && (catal != NULL) &&
 			(xmlCatalogIsEmpty(catal))) {
 			super = xmlLoadSGMLSuperCatalog(
@@ -434,10 +449,12 @@ int main(int argc, char **argv) {
 			if (super != NULL) {
 			    ret = xmlACatalogRemove(super,
 				    BAD_CAST argv[i + 1]);
-			    if (ret < 0)
+			    if (ret < 0) {
 				fprintf(stderr,
-					"Failed to removed entry from %s\n",
+					"Failed to remove entry from %s\n",
 					XML_SGML_DEFAULT_CATALOG);
+				exit_value = 1;
+			    }
 			}
 		    }
 		}
@@ -451,6 +468,7 @@ int main(int argc, char **argv) {
 			if (out == NULL) {
 			    fprintf(stderr, "could not open %s for saving\n",
 				    argv[i + 1]);
+			    exit_value = 2;
 			    noout = 0;
 			} else {
 			    xmlACatalogDump(catal, out);
@@ -466,6 +484,7 @@ int main(int argc, char **argv) {
 				fprintf(stderr,
 					"could not open %s for saving\n",
 					XML_SGML_DEFAULT_CATALOG);
+				exit_value = 2;
 				noout = 0;
 			    } else {
 				
@@ -488,12 +507,19 @@ int main(int argc, char **argv) {
 			    ret = xmlCatalogAdd(BAD_CAST argv[i + 1],
 						BAD_CAST argv[i + 2],
 						BAD_CAST argv[i + 3]);
-			if (ret != 0)
+			if (ret != 0) {
 			    printf("add command failed\n");
+			    exit_value = 3;
+			}
 			i += 3;
 		} else if ((!strcmp(argv[i], "-del")) ||
 		    (!strcmp(argv[i], "--del"))) {
 		    ret = xmlCatalogRemove(BAD_CAST argv[i + 1]);
+		    if (ret < 0) {
+			fprintf(stderr, "Failed to remove entry %s\n",
+				argv[i + 1]);
+			exit_value = 1;
+		    }
 		    i += 1;
 		}
 	    }
@@ -511,6 +537,7 @@ int main(int argc, char **argv) {
 		ans = xmlCatalogResolvePublic((const xmlChar *) argv[i]);
 		if (ans == NULL) {
 		    printf("No entry for PUBLIC %s\n", argv[i]);
+		    exit_value = 4;
 		} else {
 		    printf("%s\n", ans);
 		    xmlFree(ans);
@@ -520,6 +547,7 @@ int main(int argc, char **argv) {
 		ans = xmlCatalogResolveSystem((const xmlChar *) argv[i]);
 		if (ans == NULL) {
 		    printf("No entry for SYSTEM %s\n", argv[i]);
+		    exit_value = 4;
 		} else {
 		    printf("%s\n", ans);
 		    xmlFree(ans);
@@ -534,6 +562,7 @@ int main(int argc, char **argv) {
 	    out = fopen(filename, "w");
 	    if (out == NULL) {
 		fprintf(stderr, "could not open %s for saving\n", filename);
+		exit_value = 2;
 		noout = 0;
 	    } else {
 		xmlCatalogDump(out);
@@ -548,7 +577,7 @@ int main(int argc, char **argv) {
      */
     xmlCleanupParser();
     xmlMemoryDump();
-    return(0);
+    return(exit_value);
 }
 #else
 int main(int argc, char **argv) {
