@@ -524,6 +524,8 @@ xmlTextReaderValidateEntity(xmlTextReaderPtr reader) {
 	if (node->children != NULL) {
 	    node = node->children;
 	    continue;
+	} else if (node->type == XML_ELEMENT_NODE) {
+	    xmlTextReaderValidatePop(reader);
 	}
 	if (node->next != NULL) {
 	    node = node->next;
@@ -598,7 +600,11 @@ xmlTextReaderRead(xmlTextReaderPtr reader) {
 		return(-1);
 	    reader->state = XML_TEXTREADER_ELEMENT;
 	} else {
-	    reader->node = reader->ctxt->nodeTab[0];
+	    if (reader->ctxt->myDoc != NULL) {
+		reader->node = reader->ctxt->myDoc->children;
+	    }
+	    if (reader->node == NULL)
+		reader->node = reader->ctxt->nodeTab[0];
 	    reader->state = XML_TEXTREADER_ELEMENT;
 	}
 	reader->depth = 0;
@@ -632,6 +638,45 @@ get_next_node:
 	if (reader->node == NULL)
 	    goto node_end;
     }
+    /*
+     * If we are in the middle of a piece of CDATA make sure it's finished
+     * Maybe calling a function checking that a non-character() callback was
+     * received would be cleaner for the loop exit.
+     */
+    if ((oldstate == XML_TEXTREADER_ELEMENT) &&
+	(reader->ctxt->instate == XML_PARSER_CDATA_SECTION)) {
+	while ((reader->ctxt->instate == XML_PARSER_CDATA_SECTION) &&
+	       (((reader->node->content == NULL) &&
+		 (reader->node->next != NULL) &&
+		 (reader->node->next->type == XML_CDATA_SECTION_NODE) &&
+		 (reader->node->next->next == NULL) &&
+		 (reader->node->parent->next == NULL)) ||
+	        ((reader->node->children != NULL) &&
+		 (reader->node->children->type == XML_CDATA_SECTION_NODE) &&
+		 (reader->node->children->next == NULL) &&
+		 (reader->node->children->next == NULL)))) {
+	    val = xmlTextReaderPushData(reader);
+	    if (val < 0)
+		return(-1);
+	}
+    }
+    if ((oldstate == XML_TEXTREADER_ELEMENT) &&
+	(reader->ctxt->instate == XML_PARSER_CONTENT)) {
+	while ((reader->ctxt->instate == XML_PARSER_CONTENT) &&
+	       (((reader->node->content == NULL) &&
+		 (reader->node->next != NULL) &&
+		 (reader->node->next->type == XML_TEXT_NODE) &&
+		 (reader->node->next->next == NULL) &&
+		 (reader->node->parent->next == NULL)) ||
+	        ((reader->node->children != NULL) &&
+		 (reader->node->children->type == XML_TEXT_NODE) &&
+		 (reader->node->children->next == NULL) &&
+		 (reader->node->children->next == NULL)))) {
+	    val = xmlTextReaderPushData(reader);
+	    if (val < 0)
+		return(-1);
+	}
+    }
     if (oldstate != XML_TEXTREADER_BACKTRACK) {
 	if ((reader->node->children != NULL) &&
 	    (reader->node->type != XML_ENTITY_REF_NODE) &&
@@ -1865,7 +1910,7 @@ xmlTextReaderPrefix(xmlTextReaderPtr reader) {
     if ((node->type != XML_ELEMENT_NODE) &&
 	(node->type != XML_ATTRIBUTE_NODE))
 	return(NULL);
-    if ((node->ns != NULL) || (node->ns->prefix != NULL))
+    if ((node->ns != NULL) && (node->ns->prefix != NULL))
 	return(xmlStrdup(node->ns->prefix));
     return(NULL);
 }
@@ -1991,7 +2036,7 @@ xmlTextReaderHasValue(xmlTextReaderPtr reader) {
         case XML_COMMENT_NODE:
 	    return(1);
 	default:
-	    return(0);
+	    break;
     }
     return(0);
 }
@@ -2037,7 +2082,7 @@ xmlTextReaderValue(xmlTextReaderPtr reader) {
             if (node->content != NULL)
                 return (xmlStrdup(node->content));
 	default:
-	    return(NULL);
+	    break;
     }
     return(NULL);
 }
@@ -2347,7 +2392,7 @@ xmlTextReaderLocatorBaseURI(xmlTextReaderLocatorPtr locator) {
 	if ((input->filename == NULL) && (ctx->inputNr > 1))
 	    input = ctx->inputTab[ctx->inputNr - 2];
 	if (input != NULL) {
-	    ret = xmlStrdup(input->filename);
+	    ret = xmlStrdup(BAD_CAST input->filename);
 	} 
 	else {
 	    ret = NULL;
